@@ -20,10 +20,9 @@
 * along with GNU LibreJS.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-var acorn_base = require("acorn");
-var acorn = require('acorn/dist/acorn_loose');
+var acorn = require('acorn');
+var acornLoose = require('acorn-loose');
 var jssha = require('jssha');
-var walk = require("acorn/dist/walk");
 var legacy_license_lib = require("./legacy_license_check.js");
 var {ResponseProcessor} = require("./bg/ResponseProcessor");
 var {Storage, ListStore} = require("./common/Storage");
@@ -62,29 +61,6 @@ function hash(source){
 	return shaObj.getHash("HEX");
 }
 
-
-// the list of all available event attributes
-var intrinsic_events = [
-    "onload",
-    "onunload",
-    "onclick",
-    "ondblclick",
-    "onmousedown",
-    "onmouseup",
-    "onmouseovr",
-    "onmousemove",
-    "onmouseout",
-    "onfocus",
-    "onblur",
-    "onkeypress",
-    "onkeydown",
-    "onkeyup",
-    "onsubmit",
-    "onreset",
-    "onselect",
-    "onchange"
-];
-
 /*
 	NONTRIVIAL THINGS:
 	- Fetch
@@ -340,13 +316,18 @@ function get_domain(url){
 *
 */
 var portFromCS;
-function connected(p) {
-	if(p["name"] == "contact_finder"){
+async function connected(p) {
+	if(p.name === "contact_finder"){
+		// style the contact finder panel
+		await browser.tabs.insertCSS(p.sender.tab.id, {
+			file: "/content/dialog.css",
+			cssOrigin: "user",
+			matchAboutBlank: true,
+			allFrames: true
+		});
+
 		// Send a message back with the relevant settings
-		function cb(items){
-			p.postMessage(items);
-		}
-		browser.storage.local.get(cb);
+		p.postMessage(await browser.storage.local.get(["prefs_subject", "prefs_body"]));
 		return;
 	}
 	p.onMessage.addListener(async function(m) {
@@ -373,7 +354,7 @@ function connected(p) {
 		// invoke_contact_finder
 		if(m["invoke_contact_finder"] !== undefined){
 			contact_finder = true;
-			inject_contact_finder();
+			await injectContactFinder();
 		}
 		// a debug feature (maybe give the user an option to do this?)
 		if(m["deletelocalstorage"] !== undefined){
@@ -462,7 +443,7 @@ function full_evaluate(script){
 			return [true,"Harmless null script"];
 		}
 
-		var ast = acorn.parse_dammit(script).body[0];
+		var ast = acornLoose.parse(script).body[0];
 
 		var flag = false;
 		var amtloops = 0;
@@ -470,7 +451,7 @@ function full_evaluate(script){
 		var loopkeys = {"for":true,"if":true,"while":true,"switch":true};
 		var operators = {"||":true,"&&":true,"=":true,"==":true,"++":true,"--":true,"+=":true,"-=":true,"*":true};
 		try{
-			var tokens = acorn_base.tokenizer(script);
+			var tokens = acorn.tokenizer(script);
 		}catch(e){
 			console.warn("Tokenizer could not be initiated (probably invalid code)");
 			return [false,"Tokenizer could not be initiated (probably invalid code)"];
@@ -512,7 +493,8 @@ function full_evaluate(script){
 			return script.charAt(end+i) == "[";
 		}
 		var error_count = 0;
-		while(toke !== undefined && toke.type != acorn_base.tokTypes.eof){
+		var defines_functions = false;
+		while(toke !== undefined && toke.type != acorn.tokTypes.eof){
 			if(toke.type.keyword !== undefined){
 				//dbg_print("Keyword:");
 				//dbg_print(toke);
@@ -520,10 +502,8 @@ function full_evaluate(script){
 				// This type of loop detection ignores functional loop alternatives and ternary operators
 
 				if(toke.type.keyword == "function"){
-					dbg_print("%c NONTRIVIAL: Function declaration.","color:red");
-					if(DEBUG == false){
-						return [false,"NONTRIVIAL: Function declaration."];
-					}
+					dbg_print("%c NOTICE: Function declaration.","color:green");
+					defines_functions = true;
 				}
 
 				if(loopkeys[toke.type.keyword] !== undefined){
@@ -554,13 +534,6 @@ function full_evaluate(script){
 						}
 					}
 				}else if(status === undefined){// is the identifier user defined?
-					// Are arguments being passed to a user defined variable?
-					if(being_called(toke.end)){
-						dbg_print("%c NONTRIVIAL: User defined variable '"+toke.value+"' called as function","color:red");
-						if(DEBUG == false){
-							return [false,"NONTRIVIAL: User defined variable '"+toke.value+"' called as function"];
-						}
-					}
 					// Is there bracket suffix notation?
 					if(is_bsn(toke.end)){
 						dbg_print("%c NONTRIVIAL: Bracket suffix notation on variable '"+toke.value+"'","color:red");
@@ -582,7 +555,10 @@ function full_evaluate(script){
 		}
 
 		dbg_print("%cAppears to be trivial.","color:green;");
-		return [true,"Script appears to be trivial."];
+		if (defines_functions === true)
+			return [true,"Script appears to be trivial but defines functions."];
+		else
+			return [true,"Script appears to be trivial."];
 }
 
 
@@ -595,6 +571,8 @@ function full_evaluate(script){
 *	This can only determine if a script is bad, not if it's good
 *
 *	If it passes the intitial pass, it runs the full pass and returns the result
+
+*	It returns an array of [flag (boolean, false if "bad"), reason (string, human readable report)]
 *
 */
 function evaluate(script,name){
@@ -632,27 +610,32 @@ function evaluate(script,name){
 		return [flag,reason];
 	}
 
-	var final = full_evaluate(script);
-//	final[1] = final[1] + "<br>";
-	return final;
+	return full_evaluate(script);
 }
 
 
 
-function license_valid(matches){
-	if(matches.length != 4){
-		return [false, "malformed or unrecognized license tag"];
+function validateLicense(matches) {
+	if (!(Array.isArray(matches) && matches.length >= 4)){
+		return [false, "Malformed or unrecognized license tag."];
 	}
-	if(matches[1] != "@license"){
-		return [false, "malformed or unrecognized license tag"];
+	let [all, tag, link, id] = matches;
+	let license = null;
+	if (licenses[id])
+		license = licenses[id];
+	for (let key in licenses){
+		if (licenses[key]["Magnet link"] === link)
+			license = licenses[key];
+		if (licenses[key]["URL"] === link)
+			license = licenses[key];
 	}
-	if(licenses[matches[3]] === undefined){
-		return [false, "malformed or unrecognized license tag"];
+	if(!license){
+		return [false, `Unrecognized license "${id}"`];
 	}
-	if(licenses[matches[3]]["Magnet link"] != matches[2]){
-		return [false, "malformed or unrecognized license tag"];
+	if (!(license["Magnet link"] === link || license["URL"] === link)){
+		return [false, `License magnet link does not match for "${id}".`];
 	}
-	return [true,"Recognized license as '"+matches[3]+"'<br>"];
+	return [true, `Recognized license: "${id}".`];
 }
 /**
 *
@@ -665,91 +648,92 @@ function license_valid(matches){
 *		reason text
 *	]
 */
-function license_read(script_src, name, external = false){
+function license_read(scriptSrc, name, external = false){
 
-	var reason_text = "";
+	let license = legacy_license_lib.check(scriptSrc);
+	if (license){
+		return [true, scriptSrc, `Licensed under: ${license}`];
+	}
+	if (listManager.builtInHashes.has(hash(scriptSrc))){
+		return [true, scriptSrc, "Common script known to be free software."];
+	}
 
-	var edited_src = "";
-	var unedited_src = script_src;
-	var nontrivial_status;
-	var parts_denied = false;
-	var parts_accepted = false;
-	var license = legacy_license_lib.check(script_src);
-	if(license != false){
-		return [true,script_src,"Licensed under: "+license];
-	}
-	if (listManager.builtInHashes.has(hash(script_src))){
-		return [true,script_src,"Common script known to be free software."];
-	}
-	while(true){ // TODO: refactor me
-		// TODO: support multiline comments
-		var matches = /\/[\/\*]\s*?(@license)\s([\S]+)\s([\S]+$)/gm.exec(unedited_src);
-		var empty = /[^\s]/gm.exec(unedited_src);
-		if(empty == null){
-			return [true,edited_src,reason_text];
+	let editedSrc = "";
+	let uneditedSrc = scriptSrc.trim();
+	let reason = uneditedSrc ? "" : "Empty source.";
+	let partsDenied = false;
+	let partsAccepted = false;
+
+	function checkTriviality(s) {
+		if (!s.trim()) {
+			return true; // empty, ignore it
 		}
-		if(matches == null){
-			if (external)
-				return [false,edited_src,"External script with no known license."];
+		let [trivial, message] = external ?
+			[false, "External script with no known license"]
+			: evaluate(s, name);
+		if (trivial) {
+			partsAccepted = true;
+			editedSrc += s;
+		} else {
+			partsDenied = true;
+			if (s.startsWith("javascript:"))
+				editedSrc += `# LIBREJS BLOCKED: ${message}`;
 			else
-				nontrivial_status = evaluate(unedited_src,name);
-			if(nontrivial_status[0] == true){
-				parts_accepted = true;
-				edited_src += unedited_src;
-			} else{
-				parts_denied = true;
-				edited_src += "\n/*\nLIBREJS BLOCKED:"+nontrivial_status[1]+"\n*/\n";
-			}
-			reason_text += "\n" + nontrivial_status[1];
-
-			if(parts_denied == true && parts_accepted == true){
-				reason_text = "Script was determined partly non-trivial after editing. (check source for details)\n"+reason_text;
-			}
-			if(parts_denied == true && parts_accepted == false){
-				return [false,edited_src,reason_text];
-			}
-			else return [true,edited_src,reason_text];
-
+				editedSrc += `/*\nLIBREJS BLOCKED: ${message}\n*/`;
 		}
-		// sponge
-		dbg_print("undedited_src:");
-		dbg_print(unedited_src);
-		dbg_print(matches);
-		dbg_print("chopping at " + matches["index"] + ".");
-		var before = unedited_src.substring(0,matches["index"]);
-		// sponge
-		dbg_print("before:");
-		dbg_print(before);
-		if (external)
-			nontrivial_status = [true, "External script with no known license"]
-		else
-			nontrivial_status = evaluate(before,name);
-		if(nontrivial_status[0] == true){
-			parts_accepted = true;
-			edited_src += before;
-		} else{
-			parts_denied = true;
-			edited_src += "\n/*\nLIBREJS BLOCKED:"+nontrivial_status[1]+"\n*/\n";
-		}
-		unedited_src = unedited_src.substr(matches["index"],unedited_src.length);
-		// TODO: support multiline comments
-		var matches_end = /\/\/\s*?(@license-end)/gm.exec(unedited_src);
-		if(matches_end == null){
-			dbg_print("ERROR: @license with no @license-end");
-			return [false,"\n/*\n ERROR: @license with no @license-end \n*/\n","ERROR: @license with no @license-end"];
-		}
-		var endtag_end_index = matches_end["index"]+matches_end[0].length;
-		var license_res = license_valid(matches);
-		if(license_res[0] == true){
-			edited_src =  edited_src + unedited_src.substr(0,endtag_end_index);
-			reason_text += "\n" + license_res[1];
-		} else{
-			edited_src = edited_src + "\n/*\n"+license_res[1]+"\n*/\n";
-			reason_text += "\n" + license_res[1];
-		}
-		// trim off everything we just evaluated
-		unedited_src = unedited_src.substr(endtag_end_index,unedited_src.length);
+		reason += `\n${message}`;
+		return trivial;
 	}
+
+	while (uneditedSrc) {
+		let openingMatch = /\/[\/\*]\s*?(@license)\s+(\S+)\s+(\S+)\s*$/mi.exec(uneditedSrc);
+		if (!openingMatch) { // no license found, check for triviality
+			checkTriviality(uneditedSrc);
+			break;
+		}
+
+		let openingIndex = openingMatch.index;
+		if (openingIndex) {
+			// let's check the triviality of the code before the license tag, if any
+			checkTriviality(uneditedSrc.substring(0, openingIndex));
+		}
+		// let's check the actual license
+		uneditedSrc = uneditedSrc.substring(openingIndex);
+
+		let closureMatch = /\/([*/])\s*@license-end\b[^*/\n]*/i.exec(uneditedSrc);
+		if (!closureMatch) {
+			let msg = "ERROR: @license with no @license-end";
+			return [false, `\n/*\n ${msg} \n*/\n`, msg];
+		}
+
+		let closureEndIndex = closureMatch.index + closureMatch[0].length;
+		let commentEndOffset = uneditedSrc.substring(closureEndIndex).indexOf(closureMatch[1] === "*" ? "*/" : "\n");
+		if (commentEndOffset !== -1) {
+			closureEndIndex += commentEndOffset;
+		}
+
+		let [licenseOK, message] = validateLicense(openingMatch);
+		if(licenseOK) {
+			editedSrc += uneditedSrc.substr(0, closureEndIndex);
+			partsAccepted = true;
+		} else {
+			editedSrc += `\n/*\n${message}\n*/\n`;
+			partsDenied = true;
+		}
+		reason += `\n${message}`;
+
+		// trim off everything we just evaluated
+		uneditedSrc = uneditedSrc.substring(closureEndIndex).trim();
+	}
+
+	if(partsDenied) {
+		if (partsAccepted) {
+			reason = `Some parts of the script have been disabled (check the source for details).\n^--- ${reason}`;
+		}
+		return [false, editedSrc, reason];
+	}
+
+	return [true, scriptSrc, reason];
 }
 
 /* *********************************************************************************************** */
@@ -775,7 +759,10 @@ async function get_script(response, url, tabId = -1, whitelisted = false, index
 				: "Address whitelisted by user";
 			addReportEntry(tabId, url, {"whitelisted": [url, reason], url});
 		}
-		return result(`/* LibreJS: script whitelisted by user preference. */\n${response}`);
+		if (response.startsWith("javascript:"))
+			return result(response);
+		else
+			return result(`/* LibreJS: script whitelisted by user preference. */\n${response}`);
 	}
 
 	let [verdict, editedSource, reason] = license_read(response, scriptName, index === -2);
@@ -792,10 +779,20 @@ async function get_script(response, url, tabId = -1, whitelisted = false, index
 	let scriptSource = verdict ? response : editedSource;
 	switch(category) {
 		case "blacklisted":
+			if (response.startsWith("javascript:"))
+				return result(`# LibreJS: script ${category} by user.`);
+			else
+				return result(`/* LibreJS: script ${category} by user. */`);
 		case "whitelisted":
-			return result(`/* LibreJS: script ${category} by user. */\n${scriptSource}`);
+			if (response.startsWith("javascript:"))
+				return result(scriptSource);
+			else
+				return result(`/* LibreJS: script ${category} by user. */\n${scriptSource}`);
 		default:
-			return result(`/* LibreJS: script ${category}. */\n${scriptSource}`);
+                        if (response.startsWith("javascript:"))
+				return result(scriptSource);
+			else
+				return result(`/* LibreJS: script ${category}. */\n${scriptSource}`);
 	}
 }
 
@@ -858,6 +855,7 @@ var ResponseHandler = {
 		let {request} = response;
 		let {url, type, tabId, frameId, documentUrl} = request;
 
+		let fullUrl = url;
 		url = ListStore.urlItem(url);
 		let site = ListStore.siteItem(url);
 
@@ -887,11 +885,11 @@ var ResponseHandler = {
 						"whitelisted": [url, whitelistedSite ? `User whitelisted ${site}` : "Whitelisted by user"]});
 					return ResponseProcessor.ACCEPT;
 				} else {
-					let scriptInfo = await ExternalLicenses.check({url, tabId, frameId, documentUrl});
+					let scriptInfo = await ExternalLicenses.check({url: fullUrl, tabId, frameId, documentUrl});
 					if (scriptInfo) {
 						let verdict, ret;
 						let msg = scriptInfo.toString();
-						if (scriptInfo.allFree) {
+						if (scriptInfo.free) {
 							verdict = "accepted";
 							ret = ResponseProcessor.ACCEPT;
 						} else {
@@ -1054,20 +1052,21 @@ async function editHtml(html, documentUrl, tabId, frameId, whitelisted){
 		let modified = false;
 
 		// Deal with intrinsic events
+		let intrinsecindex = 0;
 		for (let element of html_doc.all) {
-			let attributes = element.attributes;
-			for (let event of intrinsic_events) {
-				if (event in attributes) {
-					let attr = attributes[event];
+			for (let attr of element.attributes){
+				if (attr.name.startsWith("on") || (attr.name === "href" && attr.value.startsWith("javascript:"))){
+					intrinsecindex++;
 					try {
-						let edited = await get_script(attr.value, `Intrinsic event [${event}]`);
-					  if (edited) {
-							let value = edited[0];
-							if (value !== attr.value) {
-								modified = true;
-								attr.value = value;
+						let url = `${documentUrl}# Intrinsic event ${intrinsecindex} [${attr.name}]`;
+						let edited = await get_script(attr.value, url, tabId, whitelist.contains(url));
+							if (edited) {
+								let value = edited;
+								if (value !== attr.value) {
+									modified = true;
+									attr.value = value;
+								}
 							}
-						}
 					} catch (e) {
 						console.error(e);
 					}
@@ -1078,6 +1077,7 @@ async function editHtml(html, documentUrl, tabId, frameId, whitelisted){
 		let modifiedInline = false;
 		for(let i = 0, len = scripts.length; i < len; i++) {
 			let script = scripts[i];
+			let url = `${documentUrl}# script ${i}`;
 			if (!script.src && !(script.type && script.type !== "text/javascript")) {
 				let edited = await get_script(script.textContent, url, tabId, whitelisted, i);
 				if (edited) {
@@ -1089,11 +1089,11 @@ async function editHtml(html, documentUrl, tabId, frameId, whitelisted){
 					}
 				}
 			}
-			if (modified) {
-				return modifiedInline
-					? await remove_noscripts(html_doc)
-					: doc2HTML(html_doc);
-			}
+		}
+		if (modified) {
+			return modifiedInline
+				? await remove_noscripts(html_doc)
+				: doc2HTML(html_doc);
 		}
 	}
 	return null;
@@ -1121,11 +1121,35 @@ var listManager = new ListManager(whitelist, blacklist,
 			.map(script => script.hash)
 	);
 
+
+async function initDefaults() {
+	let defaults = {
+		pref_subject: "Issues with Javascript on your website",
+		pref_body: `Please consider using a free license for the Javascript on your website.
+
+[Message generated by LibreJS. See https://www.gnu.org/software/librejs/ for more information]
+`
+	};
+	let keys = Object.keys(defaults);
+	let prefs = await browser.storage.local.get(keys);
+	let changed = false;
+	for (let k of keys) {
+		if (!(k in prefs)) {
+			prefs[k] = defaults[k];
+			changed = true;
+		}
+	}
+	if (changed) {
+		await browser.storage.local.set(prefs);
+	}
+}
+
 /**
 *	Initializes various add-on functions
 *	only meant to be called once when the script starts
 */
-async function init_addon(){
+async function init_addon() {
+	await initDefaults();
 	await whitelist.load();
 	browser.runtime.onConnect.addListener(connected);
 	browser.storage.onChanged.addListener(options_listener);
@@ -1149,17 +1173,32 @@ async function init_addon(){
 	ResponseProcessor.install(ResponseHandler);
 
 	legacy_license_lib.init();
+
+
+	let Test = require("./common/Test");
+	if (Test.getURL()) {
+		// export testable functions to the global scope
+		this.LibreJS = {
+			editHtml,
+			handle_script,
+			ExternalLicenses,
+		};
+		// create or focus the autotest tab if it's a debugging session
+		if ((await browser.management.getSelf()).installType === "development") {
+			Test.getTab(true);
+		}
+	}
 }
 
 
 /**
 *	Loads the contact finder on the given tab ID.
 */
-function inject_contact_finder(tab_id){
-	function executed(result) {
-	  dbg_print("[TABID:"+tab_id+"]"+"finished executing contact finder: " + result);
-	}
-	var executing = browser.tabs.executeScript(tab_id, {file: "/contact_finder.js"}, executed);
+async function injectContactFinder(tabId){
+	await Promise.all([
+		browser.tabs.insertCSS(tabId, {file: "/content/overlay.css", cssOrigin: "user"}),
+		browser.tabs.executeScript(tabId, {file: "/content/contactFinder.js"}),
+ ]);
 }
 
 init_addon();
